from __future__ import annotations
from leech_tentacle.extraction import ExtractedData
from leech_tentacle.extraction_config import ExtractionConfig


class Extractor:
    @classmethod
    def generate(cls, extraction_config: ExtractionConfig) -> Extractor:
        raise NotImplementedError()

    def perform_extraction(self, extraction_config: ExtractionConfig) -> ExtractedData:
        raise NotImplementedError()



from __future__ import annotations

from leech_tentacle.extraction import ExtractedData, ProcessedExtraction
from leech_tentacle.extraction_config import ExtractionConfig


class Processor:
    @classmethod
    def generate(cls, extraction_config: ExtractionConfig) -> Processor:
        raise NotImplementedError()

    def process(self, extraction: ExtractedData) -> ProcessedExtraction:
        raise NotImplementedError()



from algernon import AlgObject


class ExtractionConfig(AlgObject):
    def __init__(
            self,
            id_source: str,
            source_name: str,
            extraction_type: str,
            extractor_class: str,
            extractor_params: str,
            processor_class: str,
            processor_params,
            extracted_object_type: str,
            extraction_params
    ):
        self._id_source = id_source
        self._source_name = source_name
        self._extraction_type = extraction_type
        self._extractor_class = extractor_class
        self._extractor_params = extractor_params
        self._processor_class = processor_class
        self._processor_params = processor_params
        self._extracted_object_type = extracted_object_type
        self._extraction_params = extraction_params

    @classmethod
    def parse_json(cls, json_dict):
        return cls(
            id_source=json_dict['id_source'],
            source_name=json_dict['source_name'],
            extraction_type=json_dict['extraction_type'],
            extractor_class=json_dict['extractor_class'],
            extractor_params=json_dict['extractor_params'],
            processor_class=json_dict['processor_class'],
            processor_params=json_dict['processor_params'],
            extracted_object_type=json_dict['extracted_object_type'],
            extraction_params=json_dict['extraction_params']
        )

    @property
    def id_source(self):
        return self._id_source

    @property
    def source_name(self):
        return self._source_name

    @property
    def extraction_type(self) -> str:
        return self._extraction_type

    @property
    def extractor_class(self) -> str:
        return self._extractor_class

    @property
    def extractor_params(self):
        return self._extractor_params

    @property
    def processor_class(self) -> str:
        return self._processor_class

    @property
    def processor_params(self):
        return self._processor_params

    @property
    def extracted_object_type(self) -> str:
        return self._extracted_object_type

    @property
    def extraction_params(self):
        return self._extraction_params



from __future__ import annotations

from typing import TypedDict, Mapping, Union, Sequence

TemplatePrimitive = Union[
    float,
    int,
    str,
    bool,
    Sequence['TemplatePrimitive'],
    Mapping[str, 'TemplatePrimitive']
]


class RuntimeParameter(TypedDict):
    insert_path: str
    parameter_name: str


class DynamicParameter(TypedDict):
    dynamic: bool
    dynamic_type: str
    dynamic_kwargs: TemplatePrimitive


TemplateParameter = Union[TemplatePrimitive, DynamicParameter]
TemplateParameters = Mapping[str, TemplateParameter]


class ExtractionParameters(TypedDict):
    runtime_parameters: Sequence[RuntimeParameter]
    static_parameters: TemplateParameters


class ExtractorTemplate(TypedDict):
    extractor_class: str
    runtime_parameters: Sequence[RuntimeParameter]
    static_parameters: TemplateParameters


class ProcessorTemplate(TypedDict):
    processor_class: str
    run_time_parameters: Sequence[RuntimeParameter]
    static_parameters: TemplateParameters


ParameterSet = Union[
    ExtractorTemplate,
    ProcessorTemplate,
    ExtractionParameters
]


class ExtractionTemplate(TypedDict):
    extraction_type: str
    extracted_object_type: str
    extraction: ExtractionParameters
    extractor: ExtractorTemplate
    processor: ProcessorTemplate



class TentacleException(Exception):
    """ super class for exceptions specific to the tentacle.

    """
    @property
    def message(self):
        raise NotImplementedError()


class MalformedConfigException(TentacleException):
    """ exceptions to handle ExtractionConfig and SourceConfig data that does not conform to standard.

    """
    def __init__(self, malformed_component, config_data, malformation_type):
        self._malformed_component = malformed_component
        self._config_data = config_data
        self._malformation_type = malformation_type

    @property
    def message(self):
        return f'could not generate {self._malformed_component} from {self._config_data}, {self._malformation_type}'


class UnknownObjectException(TentacleException):
    """ exceptions to handle requests for objects or configurations that do not exist in the tentacle.

    """
    def __init__(self, object_type: str, object_value: str):
        self._object_type = object_type
        self._object_value = object_value

    @property
    def message(self):
        return f'{self._object_type} {self._object_value} is not registered with the tentacle'


class UnknownExtractionTypeException(UnknownObjectException):
    """ raised when an unregistered ExtractionType is invoked.

    """
    def __init__(self, extraction_type: str):
        super(UnknownExtractionTypeException, self).__init__('ExtractionType', extraction_type)


class UnknownExtractorException(UnknownObjectException):
    """ raised when an extraction is attempted with an unknown Extractor class.

    """
    def __init__(self, extractor_class: str):
        super(UnknownExtractorException, self).__init__('Extractor', extractor_class)


class UnknownProcessorException(UnknownObjectException):
    """ raised when an extraction is attempted with an unknown Processor class.

    """
    def __init__(self, processor_class: str):
        super(UnknownProcessorException, self).__init__('Processor', processor_class)


class UnknownIdSourceException(UnknownObjectException):
    """ raised when an extraction is attempted for an unknown IdSource.

    """
    def __init__(self, id_source: str):
        super(UnknownIdSourceException, self).__init__('IdSource', id_source)



from typing import Iterable

from algernon import AlgObject

from leech_tentacle.extraction_config import ExtractionConfig


class ExtractedData(AlgObject):
    def __init__(
            self,
            extraction_type: str,
            capture_timestamp: float,
            object_type: str,
            id_source: str,
            source_name: str,
            extracted_data: object
    ):
        self._extraction_type = extraction_type
        self._capture_timestamp = capture_timestamp
        self._object_type = object_type
        self._id_source = id_source
        self._source_name = source_name
        self._extracted_data = extracted_data

    @classmethod
    def parse_json(cls, json_dict):
        return cls(
            capture_timestamp=json_dict['capture_timestamp'],
            extraction_type=json_dict['extraction_type'],
            object_type=json_dict['object_type'],
            id_source=json_dict['id_source'],
            source_name=json_dict['source_name'],
            extracted_data=json_dict['extracted_data']
        )

    @classmethod
    def from_extraction_config(
            cls,
            extraction_config: ExtractionConfig,
            capture_timestamp: float,
            extracted_data: object
    ):
        return cls(
            extraction_type=extraction_config.extraction_type,
            object_type=extraction_config.extracted_object_type,
            id_source=extraction_config.id_source,
            source_name=extraction_config.source_name,
            extracted_data=extracted_data,
            capture_timestamp=capture_timestamp
        )

    @property
    def extraction_type(self) -> str:
        return self._extraction_type

    @property
    def object_type(self) -> str:
        return self._object_type

    @property
    def id_source(self) -> str:
        return self._id_source

    @property
    def source_name(self) -> str:
        return self._source_name

    @property
    def extracted_data(self) -> object:
        return self._extracted_data

    @property
    def capture_timestamp(self) -> float:
        return self._capture_timestamp


class ProcessedExtraction(AlgObject):
    def __init__(
            self,
            id_source: str,
            source_name: str,
            capture_timestamp: float,
            extracted_object_type: str,
            processed_data: Iterable[object]
    ):
        self._id_source = id_source
        self._source_name = source_name
        self._capture_timestamp = capture_timestamp
        self._extracted_object_type = extracted_object_type
        self._processed_data = processed_data

    @classmethod
    def parse_json(cls, json_dict):
        return cls(
            id_source=json_dict['id_source'],
            source_name=json_dict['source_name'],
            capture_timestamp=json_dict['capture_timestamp'],
            extracted_object_type=json_dict['extracted_object_type'],
            processed_data=json_dict['processed_data']
        )

    @property
    def id_source(self) -> str:
        return self._id_source

    @property
    def source_name(self) -> str:
        return self._source_name

    @property
    def capture_timestamp(self) -> float:
        return self._capture_timestamp

    @property
    def extracted_object_type(self) -> str:
        return self._extracted_object_type

    @property
    def processed_data(self) -> Iterable[object]:
        return self._processed_data


class DataAsset:
    def __init__(self, id_source: str, source_name: str, capture_timestamp: float, asset_type: str, asset_data):
        self._id_source = id_source
        self._source_name = source_name
        self._capture_timestamp = capture_timestamp
        self._asset_type = asset_type
        self._asset_data = asset_data

    @property
    def id_source(self):
        return self._id_source

    @property
    def source_name(self):
        return self._source_name

    @property
    def capture_timestamp(self):
        return self._capture_timestamp

    @property
    def asset_type(self):
        return self._asset_type

    @property
    def asset_data(self):
        return self._asset_data
